from typing import ClassVar, Optional

from grpc.aio import insecure_channel
from pydantic import BaseModel, validate_call

from .api_constants import (
    SUCCESS_CODE,
    BasketRequestMessages,
    BasketRPCMethodNames,
    ControllerStubs,
    TradeRequestMessages,
    TradeRPCMethodNames,
)


class ValidationError(Exception):
    pass


class ResponseCodeError(Exception):
    pass


class Core(BaseModel):
    """
    Configure the host and port of the search server.

    Parameters:
        host (str): The host address of the search server.
        port (int): The port number of the search server.
    """

    host: str = "127.0.0.1"
    port: str | int = "50051"

    __instance: ClassVar = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    async def perform_operation(
        self, rpc_method_name, request_message, controller_stub
    ):
        """Perform the operation.

        Parameters:
            rpc_method_name (str): The name of the RPC method (e.g., `Create`).
            request_message (gRPC message): The request message instance.
        """
        async with insecure_channel(f"{self.host}:{self.port}") as channel:
            controller_stub.channel = channel
            controller_stub.ready()
            RPCMethod = getattr(controller_stub, rpc_method_name, None)

            if RPCMethod:
                response = await RPCMethod(request_message)

                # Clear the request message for future use
                request_message.Clear()

                # Raise error if response is not successful
                if response.code != SUCCESS_CODE:
                    raise ResponseCodeError(response.code)

                return response
            else:
                raise ValidationError("Invalid RPC method name.")


class Basket(Core):
    @validate_call
    async def create(
        self, owner_identifier: str, basket_identifier: Optional[str] = None
    ):
        """Create a new basket instance."""
        # Assign attributes to the request message
        BasketRequestMessages.CREATE.owner_identifier = owner_identifier
        if basket_identifier:
            BasketRequestMessages.CREATE.basket_identifier = basket_identifier

        # Perform the create operation
        return await self.perform_operation(
            BasketRPCMethodNames.CREATE,  # The name of the RPC method
            BasketRequestMessages.CREATE,  # The request message
            ControllerStubs.BASKET,  # The controller stub
        )

    @validate_call
    async def retrieve(self, basket_identifier: str):
        """Retrieve a basket instance."""
        # Assign attributes to the request message
        BasketRequestMessages.RTRIEVE.basket_identifier = basket_identifier

        # Perform the create operation
        return await self.perform_operation(
            BasketRPCMethodNames.RTRIEVE,  # The name of the RPC method
            BasketRequestMessages.RTRIEVE,  # The request message
            ControllerStubs.BASKET,  # The controller stub
        )

    @validate_call
    async def add_item(
        self,
        item_type: str,
        item_identifier: str,
        discount_token: str,
        discount_amount: int,
        price: int,
        extras: dict,
        basket_identifier: str,
    ):
        """Add item to an existing basket."""
        # Set the keyword attribute in the request message
        BasketRequestMessages.ADD_ITEM.item_type = item_type
        BasketRequestMessages.ADD_ITEM.item_identifier = item_identifier
        BasketRequestMessages.ADD_ITEM.discount_token = discount_token
        BasketRequestMessages.ADD_ITEM.discount_amount = discount_amount
        BasketRequestMessages.ADD_ITEM.price = price
        BasketRequestMessages.ADD_ITEM.extras.update(extras)
        BasketRequestMessages.ADD_ITEM.basket_identifier = basket_identifier

        # Perform the search operation
        return await self.perform_operation(
            BasketRPCMethodNames.ADD_ITEM,  # The name of the RPC method for searching
            BasketRequestMessages.ADD_ITEM,  # The request message for searching
            ControllerStubs.BASKET,  # The controller stub for searching
        )

    @validate_call
    async def remove_item(
        self, item_identifier: str, basket_identifier: str, price: int, extras: dict
    ):
        """Remove an item from an existing basket."""
        # Set the keyword attribute in the request message
        BasketRequestMessages.REMOVE_ITEM.item_identifier = item_identifier
        BasketRequestMessages.REMOVE_ITEM.basket_identifier = basket_identifier
        BasketRequestMessages.REMOVE_ITEM.price = price
        BasketRequestMessages.REMOVE_ITEM.extras.update(extras)

        # Perform the search operation
        return await self.perform_operation(
            BasketRPCMethodNames.REMOVE_ITEM,  # The name of the RPC method for searching
            BasketRequestMessages.REMOVE_ITEM,  # The request message for searching
            ControllerStubs.BASKET,  # The controller stub for searching
        )


class Trade(Core):
    @validate_call
    async def start(self, basket_identifier: str, runner: str, extras: dict):
        """Start the trade process (paying a basket)."""

        # Set the keyword attribute in the START request message
        TradeRequestMessages.TRANSACTION.runner = runner
        TradeRequestMessages.TRANSACTION.extras.update(extras)
        TradeRequestMessages.START.basket_identifier = basket_identifier
        TradeRequestMessages.START.trade_transactions.append(
            TradeRequestMessages.TRANSACTION
        )

        # Perform the search operation
        response = await self.perform_operation(
            TradeRPCMethodNames.START, TradeRequestMessages.START, ControllerStubs.TRADE
        )

        # Clear the TRANSACTION message for future use
        TradeRequestMessages.TRANSACTION.Clear()

        return response
